@@ -113,6 +113,9 @@ FLAGS = flags.FLAGS
 
 _TEXT_OR_BINARY_TYPES = (six.text_type, six.binary_type)
 
+# Suppress surplus entries in AssertionError stack traces.
+__unittest = True  # pylint: disable=invalid-name
+
 
 def expectedFailureIf(condition, reason):  # pylint: disable=invalid-name
   """Expects the test to fail if the run condition is True.
@@ -139,7 +142,7 @@ class TempFileCleanup(enum.Enum):
   # Always cleanup temp files when the test completes.
   ALWAYS = 'always'
   # Only cleanup temp file if the test passes. This allows easier inspection
-  # of tempfile contents on test failure. FLAGS.test_tempdir determines
+  # of tempfile contents on test failure. absltest.TEST_TMPDIR.value determines
   # where tempfiles are created.
   SUCCESS = 'success'
   # Never cleanup temp files.
@@ -207,7 +210,12 @@ def _get_default_randomize_ordering_seed():
     ValueError: Raised when the flag or env value is not one of the options
         above.
   """
-  randomize = FLAGS.test_randomize_ordering_seed
+  if FLAGS['test_randomize_ordering_seed'].present:
+    randomize = FLAGS.test_randomize_ordering_seed
+  elif 'TEST_RANDOMIZE_ORDERING_SEED' in os.environ:
+    randomize = os.environ['TEST_RANDOMIZE_ORDERING_SEED']
+  else:
+    randomize = ''
   if not randomize:
     return 0
   if randomize == 'random':
@@ -224,28 +232,34 @@ def _get_default_randomize_ordering_seed():
       'Unknown test randomization seed value: {}'.format(randomize))
 
 
-flags.DEFINE_integer('test_random_seed', _get_default_test_random_seed(),
-                     'Random seed for testing. Some test frameworks may '
-                     'change the default value of this flag between runs, so '
-                     'it is not appropriate for seeding probabilistic tests.',
-                     allow_override_cpp=True)
-flags.DEFINE_string('test_srcdir',
-                    get_default_test_srcdir(),
-                    'Root of directory tree where source files live',
-                    allow_override_cpp=True)
-flags.DEFINE_string('test_tmpdir', get_default_test_tmpdir(),
-                    'Directory for temporary testing files',
-                    allow_override_cpp=True)
-flags.DEFINE_string('test_randomize_ordering_seed',
-                    os.environ.get('TEST_RANDOMIZE_ORDERING_SEED', ''),
-                    'If positive, use this as a seed to randomize the '
-                    'execution order for test cases. If "random", pick a '
-                    'random seed to use. If 0 or not set, do not randomize '
-                    'test case execution order. This flag also overrides '
-                    'the TEST_RANDOMIZE_ORDERING_SEED environment variable.',
-                    allow_override_cpp=True)
-flags.DEFINE_string('xml_output_file', '',
-                    'File to store XML test results')
+TEST_SRCDIR = flags.DEFINE_string(
+    'test_srcdir',
+    get_default_test_srcdir(),
+    'Root of directory tree where source files live',
+    allow_override_cpp=True)
+TEST_TMPDIR = flags.DEFINE_string(
+    'test_tmpdir',
+    get_default_test_tmpdir(),
+    'Directory for temporary testing files',
+    allow_override_cpp=True)
+
+flags.DEFINE_integer(
+    'test_random_seed',
+    _get_default_test_random_seed(),
+    'Random seed for testing. Some test frameworks may '
+    'change the default value of this flag between runs, so '
+    'it is not appropriate for seeding probabilistic tests.',
+    allow_override_cpp=True)
+flags.DEFINE_string(
+    'test_randomize_ordering_seed',
+    '',
+    'If positive, use this as a seed to randomize the '
+    'execution order for test cases. If "random", pick a '
+    'random seed to use. If 0 or not set, do not randomize '
+    'test case execution order. This flag also overrides '
+    'the TEST_RANDOMIZE_ORDERING_SEED environment variable.',
+    allow_override_cpp=True)
+flags.DEFINE_string('xml_output_file', '', 'File to store XML test results')
 
 
 # We might need to monkey-patch TestResult so that it stops considering an
@@ -307,6 +321,10 @@ class _TempDir(object):
   """Represents a temporary directory for tests.
 
   Creation of this class is internal. Using its public methods is OK.
+
+  This class implements the `os.PathLike` interface (specifically,
+  `os.PathLike[str]`). This means, in Python 3, it can be directly passed
+  to e.g. `os.path.join()`.
   """
 
   def __init__(self, path):
@@ -317,9 +335,18 @@ class _TempDir(object):
   @property
   def full_path(self):
     # type: () -> Text
-    """The path, as a string, for the directory."""
+    """Returns the path, as a string, for the directory.
+
+    TIP: Instead of e.g. `os.path.join(temp_dir.full_path)`, you can simply
+    do `os.path.join(temp_dir)` because `__fspath__()` is implemented.
+    """
     return self._path
 
+  def __fspath__(self):
+    # type: () -> Text
+    """See os.PathLike."""
+    return self.full_path
+
   def create_file(self, file_path=None, content=None, mode='w', encoding='utf8',
                   errors='strict'):
     # type: (Optional[Text], Optional[AnyStr], Text, Text, Text) -> _TempFile
@@ -375,6 +402,10 @@ class _TempFile(object):
   """Represents a tempfile for tests.
 
   Creation of this class is internal. Using its public methods is OK.
+
+  This class implements the `os.PathLike` interface (specifically,
+  `os.PathLike[str]`). This means, in Python 3, it can be directly passed
+  to e.g. `os.path.join()`.
   """
 
   def __init__(self, path):
@@ -419,9 +450,18 @@ class _TempFile(object):
   @property
   def full_path(self):
     # type: () -> Text
-    """The path, as a string, for the file."""
+    """Returns the path, as a string, for the file.
+
+    TIP: Instead of e.g. `os.path.join(temp_file.full_path)`, you can simply
+    do `os.path.join(temp_file)` because `__fspath__()` is implemented.
+    """
     return self._path
 
+  def __fspath__(self):
+    # type: () -> Text
+    """See os.PathLike."""
+    return self.full_path
+
   def read_text(self, encoding='utf8', errors='strict'):
     # type: (Text, Text) -> Text
     """Return the contents of the file as text."""
@@ -484,7 +524,7 @@ class _TempFile(object):
                        'file in text mode'.format(mode))
     if 't' not in mode:
       mode += 't'
-    cm = self._open(mode, encoding, errors)  # type: ContextManager[TextIO]
+    cm = self._open(mode, encoding, errors)
     return cm
 
   def open_bytes(self, mode='rb'):
@@ -506,20 +546,53 @@ class _TempFile(object):
                        'file in binary mode'.format(mode))
     if 'b' not in mode:
       mode += 'b'
-    cm = self._open(mode, encoding=None, errors=None)  # type: ContextManager[BinaryIO]
+    cm = self._open(mode, encoding=None, errors=None)
     return cm
 
   # TODO(b/123775699): Once pytype supports typing.Literal, use overload and
-  # Literal to express more precise return types and remove the type comments in
-  # open_text and open_bytes.
+  # Literal to express more precise return types. The contained type is
+  # currently `Any` to avoid [bad-return-type] errors in the open_* methods.
   @contextlib.contextmanager
   def _open(self, mode, encoding='utf8', errors='strict'):
-    # type: (Text, Text, Text) -> Iterator[Union[IO[Text], IO[bytes]]]
+    # type: (Text, Text, Text) -> Iterator[Any]
     with io.open(
         self.full_path, mode=mode, encoding=encoding, errors=errors) as fp:
       yield fp
 
 
+class _method(object):
+  """A decorator that supports both instance and classmethod invocations.
+
+  Using similar semantics to the @property builtin, this decorator can augment
+  an instance method to support conditional logic when invoked on a class
+  object. This breaks support for invoking an instance method via the class
+  (e.g. Cls.method(self, ...)) but is still situationally useful.
+  """
+
+  def __init__(self, finstancemethod):
+    # type: (Callable[..., Any]) -> None
+    self._finstancemethod = finstancemethod
+    self._fclassmethod = None
+
+  def classmethod(self, fclassmethod):
+    # type: (Callable[..., Any]) -> _method
+    self._fclassmethod = classmethod(fclassmethod)
+    return self
+
+  def __doc__(self):
+    # type: () -> str
+    if getattr(self._finstancemethod, '__doc__'):
+      return self._finstancemethod.__doc__
+    elif getattr(self._fclassmethod, '__doc__'):
+      return self._fclassmethod.__doc__
+    return ''
+
+  def __get__(self, obj, type_):
+    # type: (Optional[Any], Optional[Type[Any]]) -> Callable[..., Any]
+    func = self._fclassmethod if obj is None else self._finstancemethod
+    return func.__get__(obj, type_)  # pytype: disable=attribute-error
+
+
 class TestCase(unittest3_backport.TestCase):
   """Extension of unittest.TestCase providing more power."""
 
@@ -529,25 +602,38 @@ class TestCase(unittest3_backport.TestCase):
   # tempfile module. This can be overridden at the class level, instance level,
   # or with the `cleanup` arg of `create_tempfile()` and `create_tempdir()`. See
   # `TempFileCleanup` for details on the different values.
-  tempfile_cleanup = TempFileCleanup.ALWAYS  # type: TempFileCleanup
+  # TODO(b/70517332): Remove the type comment and the disable once pytype has
+  # better support for enums.
+  tempfile_cleanup = TempFileCleanup.ALWAYS  # type: TempFileCleanup  # pytype: disable=annotation-type-mismatch
 
   maxDiff = 80 * 20
   longMessage = True
 
+  # Exit stacks for per-test and per-class scopes.
+  _exit_stack = None
+  _cls_exit_stack = None
+
   def __init__(self, *args, **kwargs):
     super(TestCase, self).__init__(*args, **kwargs)
     # This is to work around missing type stubs in unittest.pyi
     self._outcome = getattr(self, '_outcome')  # type: Optional[_OutcomeType]
-    # This is re-initialized by setUp().
-    self._exit_stack = None
 
   def setUp(self):
     super(TestCase, self).setUp()
-    # NOTE: Only Py3 contextlib has ExitStack
+    # NOTE: Only Python 3 contextlib has ExitStack
     if hasattr(contextlib, 'ExitStack'):
       self._exit_stack = contextlib.ExitStack()
       self.addCleanup(self._exit_stack.close)
 
+  @classmethod
+  def setUpClass(cls):
+    super(TestCase, cls).setUpClass()
+    # NOTE: Only Python 3 contextlib has ExitStack and only Python 3.8+ has
+    # addClassCleanup.
+    if hasattr(contextlib, 'ExitStack') and hasattr(cls, 'addClassCleanup'):
+      cls._cls_exit_stack = contextlib.ExitStack()
+      cls.addClassCleanup(cls._cls_exit_stack.close)
+
   def create_tempdir(self, name=None, cleanup=None):
     # type: (Optional[Text], Optional[TempFileCleanup]) -> _TempDir
     """Create a temporary directory specific to the test.
@@ -558,7 +644,19 @@ class TestCase(unittest3_backport.TestCase):
     This creates a named directory on disk that is isolated to this test, and
     will be properly cleaned up by the test. This avoids several pitfalls of
     creating temporary directories for test purposes, as well as makes it easier
-    to setup directories and verify their contents.
+    to setup directories and verify their contents. For example:
+
+        def test_foo(self):
+          out_dir = self.create_tempdir()
+          out_log = out_dir.create_file('output.log')
+          expected_outputs = [
+              os.path.join(out_dir, 'data-0.txt'),
+              os.path.join(out_dir, 'data-1.txt'),
+          ]
+          code_under_test(out_dir)
+          self.assertTrue(os.path.exists(expected_paths[0]))
+          self.assertTrue(os.path.exists(expected_paths[1]))
+          self.assertEqual('foo', out_log.read_text())
 
     See also: `create_tempfile()` for creating temporary files.
 
@@ -570,7 +668,8 @@ class TestCase(unittest3_backport.TestCase):
         `self.tempfile_cleanup`.
 
     Returns:
-      A _TempDir representing the created directory.
+      A _TempDir representing the created directory; see _TempDir class docs
+      for usage.
     """
     test_path = self._get_tempdir_path_test()
 
@@ -600,7 +699,13 @@ class TestCase(unittest3_backport.TestCase):
     be properly cleaned up by the test. This avoids several pitfalls of
     creating temporary files for test purposes, as well as makes it easier
     to setup files, their data, read them back, and inspect them when
-    a test fails.
+    a test fails. For example:
+
+        def test_foo(self):
+          output = self.create_tempfile()
+          code_under_test(output)
+          self.assertGreater(os.path.getsize(output), 0)
+          self.assertEqual('foo', output.read_text())
 
     NOTE: This will zero-out the file. This ensures there is no pre-existing
     state.
@@ -629,7 +734,8 @@ class TestCase(unittest3_backport.TestCase):
         `self.tempfile_cleanup`.
 
     Returns:
-      A _TempFile representing the created file.
+      A _TempFile representing the created file; see _TempFile class docs for
+      usage.
     """
     test_path = self._get_tempdir_path_test()
     tf, cleanup_path = _TempFile._create(test_path, file_path, content=content,
@@ -638,13 +744,19 @@ class TestCase(unittest3_backport.TestCase):
     self._maybe_add_temp_path_cleanup(cleanup_path, cleanup)
     return tf
 
+  @_method
   def enter_context(self, manager):
+    # type: (ContextManager[_T]) -> _T
     """Returns the CM's value after registering it with the exit stack.
 
-    Entering a context pushes it onto a stack of contexts. The context is exited
-    when the test completes. Contexts are are exited in the reverse order of
-    entering. They will always be exited, regardless of test failure/success.
-    The context stack is specific to the test being run.
+    Entering a context pushes it onto a stack of contexts. When `enter_context`
+    is called on the test instance (e.g. `self.enter_context`), the context is
+    exited after the test case's tearDown call. When called on the test class
+    (e.g. `TestCase.enter_context`), the context is exited after the test
+    class's tearDownClass call.
+
+    Contexts are are exited in the reverse order of entering. They will always
+    be exited, regardless of test failure/success.
 
     This is useful to eliminate per-test boilerplate when context managers
     are used. For example, instead of decorating every test with `@mock.patch`,
@@ -657,17 +769,25 @@ class TestCase(unittest3_backport.TestCase):
     Args:
       manager: The context manager to enter.
     """
-    # type: (ContextManager[_T]) -> _T
     if not self._exit_stack:
       raise AssertionError(
           'self._exit_stack is not set: enter_context is Py3-only; also make '
           'sure that AbslTest.setUp() is called.')
     return self._exit_stack.enter_context(manager)
 
+  @enter_context.classmethod
+  def enter_context(cls, manager):  # pylint: disable=no-self-argument
+    # type: (ContextManager[_T]) -> _T
+    if not cls._cls_exit_stack:
+      raise AssertionError(
+          'cls._cls_exit_stack is not set: cls.enter_context requires '
+          'Python 3.8+; also make sure that AbslTest.setUpClass() is called.')
+    return cls._cls_exit_stack.enter_context(manager)
+
   @classmethod
   def _get_tempdir_path_cls(cls):
     # type: () -> Text
-    return os.path.join(FLAGS.test_tmpdir, _get_qualname(cls))
+    return os.path.join(TEST_TMPDIR.value, _get_qualname(cls))
 
   def _get_tempdir_path_test(self):
     # type: () -> Text
@@ -719,7 +839,13 @@ class TestCase(unittest3_backport.TestCase):
     Returns:
       desc: A short description of a test method.
     """
-    desc = str(self)
+    desc = self.id()
+
+    # Omit the main name so that test name can be directly copy/pasted to
+    # the command line.
+    if desc.startswith('__main__.'):
+      desc = desc[len('__main__.'):]
+
     # NOTE: super() is used here instead of directly invoking
     # unittest.TestCase.shortDescription(self), because of the
     # following line that occurs later on:
@@ -1912,15 +2038,6 @@ def _quote_long_string(s):
           '----------->8\n')
 
 
-class _TestProgramManualRun(unittest.TestProgram):
-  """A TestProgram which runs the tests manually."""
-
-  def runTests(self, do_run=False):
-    """Runs the tests."""
-    if do_run:
-      unittest.TestProgram.runTests(self)
-
-
 def print_python_version():
   # type: () -> None
   # Having this in the test output logs by default helps debugging when all
@@ -1932,7 +2049,7 @@ def print_python_version():
 
 
 def main(*args, **kwargs):
-  # type: (Text, Mapping[Text, Any]) -> None
+  # type: (Text, Any) -> None
   """Executes a set of Python unit tests.
 
   Usually this function is called without arguments, so the
@@ -2071,12 +2188,93 @@ def _is_suspicious_attribute(testCaseClass, name):
   if name.startswith('Test') and len(name) > 4 and name[4].isupper():
     attr = getattr(testCaseClass, name)
     if inspect.isfunction(attr) or inspect.ismethod(attr):
-      args = inspect.getargspec(attr)
-      return (len(args.args) == 1 and args.args[0] == 'self'
-              and args.varargs is None and args.keywords is None)
+      if six.PY2:
+        args = inspect.getargspec(attr)
+        return (len(args.args) == 1 and args.args[0] == 'self'
+                and args.varargs is None and args.keywords is None)
+      else:
+        args = inspect.getfullargspec(attr)
+        return (len(args.args) == 1 and args.args[0] == 'self'
+                and args.varargs is None and args.varkw is None and
+                not args.kwonlyargs)
   return False
 
 
+def skipThisClass(reason):
+  # type: (Text) -> Callable[[_T], _T]
+  """Skip tests in the decorated TestCase, but not any of its subclasses.
+
+  This decorator indicates that this class should skip all its tests, but not
+  any of its subclasses. Useful for if you want to share testMethod or setUp
+  implementations between a number of concrete testcase classes.
+
+  Example usage, showing how you can share some common test methods between
+  subclasses. In this example, only 'BaseTest' will be marked as skipped, and
+  not RealTest or SecondRealTest:
+
+    @absltest.skipThisClass("Shared functionality")
+    class BaseTest(absltest.TestCase):
+      def test_simple_functionality(self):
+        self.assertEqual(self.system_under_test.method(), 1)
+
+    class RealTest(BaseTest):
+      def setUp(self):
+        super().setUp()
+        self.system_under_test = MakeSystem(argument)
+
+      def test_specific_behavior(self):
+        ...
+
+    class SecondRealTest(BaseTest):
+      def setUp(self):
+        super().setUp()
+        self.system_under_test = MakeSystem(other_arguments)
+
+      def test_other_behavior(self):
+        ...
+
+  Args:
+    reason: The reason we have a skip in place. For instance: 'shared test
+      methods' or 'shared assertion methods'.
+
+  Returns:
+    Decorator function that will cause a class to be skipped.
+  """
+  if isinstance(reason, type):
+    raise TypeError('Got {!r}, expected reason as string'.format(reason))
+
+  def _skip_class(test_case_class):
+    if not issubclass(test_case_class, unittest.TestCase):
+      raise TypeError(
+          'Decorating {!r}, expected TestCase subclass'.format(test_case_class))
+
+    # Only shadow the setUpClass method if it is directly defined. If it is
+    # in the parent class we invoke it via a super() call instead of holding
+    # a reference to it.
+    shadowed_setupclass = test_case_class.__dict__.get('setUpClass', None)
+
+    @classmethod
+    def replacement_setupclass(cls, *args, **kwargs):
+      # Skip this class if it is the one that was decorated with @skipThisClass
+      if cls is test_case_class:
+        raise SkipTest(reason)
+      if shadowed_setupclass:
+        # Pass along `cls` so the MRO chain doesn't break.
+        # The original method is a `classmethod` descriptor, which can't
+        # be directly called, but `__func__` has the underlying function.
+        return shadowed_setupclass.__func__(cls, *args, **kwargs)
+      else:
+        # Because there's no setUpClass() defined directly on test_case_class,
+        # we call super() ourselves to continue execution of the inheritance
+        # chain.
+        return super(test_case_class, cls).setUpClass(*args, **kwargs)
+
+    test_case_class.setUpClass = replacement_setupclass
+    return test_case_class
+
+  return _skip_class
+
+
 class TestLoader(unittest.TestLoader):
   """A test loader which supports common test features.
 
@@ -2097,10 +2295,10 @@ class TestLoader(unittest.TestLoader):
     super(TestLoader, self).__init__(*args, **kwds)
     seed = _get_default_randomize_ordering_seed()
     if seed:
-      self._seed = seed
-      self._random = random.Random(self._seed)
+      self._randomize_ordering_seed = seed
+      self._random = random.Random(self._randomize_ordering_seed)
     else:
-      self._seed = None
+      self._randomize_ordering_seed = None
       self._random = None
 
   def getTestCaseNames(self, testCaseClass):  # pylint:disable=invalid-name
@@ -2109,10 +2307,12 @@ class TestLoader(unittest.TestLoader):
       if _is_suspicious_attribute(testCaseClass, name):
         raise TypeError(TestLoader._ERROR_MSG % name)
     names = super(TestLoader, self).getTestCaseNames(testCaseClass)
-    if self._seed is not None:
-      logging.info('Randomizing test order with seed: %d', self._seed)
-      logging.info('To reproduce this order, re-run with '
-                   '--test_randomize_ordering_seed=%d', self._seed)
+    if self._randomize_ordering_seed is not None:
+      logging.info(
+          'Randomizing test order with seed: %d', self._randomize_ordering_seed)
+      logging.info(
+          'To reproduce this order, re-run with '
+          '--test_randomize_ordering_seed=%d', self._randomize_ordering_seed)
       self._random.shuffle(names)
     return names
 
@@ -2122,7 +2322,7 @@ def get_default_xml_output_filename():
   if os.environ.get('XML_OUTPUT_FILE'):
     return os.environ['XML_OUTPUT_FILE']
   elif os.environ.get('RUNNING_UNDER_TEST_DAEMON'):
-    return os.path.join(os.path.dirname(FLAGS.test_tmpdir), 'test_detail.xml')
+    return os.path.join(os.path.dirname(TEST_TMPDIR.value), 'test_detail.xml')
   elif os.environ.get('TEST_XMLOUTPUTDIR'):
     return os.path.join(
         os.environ['TEST_XMLOUTPUTDIR'],
@@ -2149,6 +2349,30 @@ def _setup_filtering(argv):
   argv[1:1] = shlex.split(test_filter)
 
 
+def _setup_test_runner_fail_fast(argv):
+  # type: (MutableSequence[Text]) -> None
+  """Implements the bazel test fail fast protocol.
+
+  The following environment variable is used in this method:
+
+    TESTBRIDGE_TEST_RUNNER_FAIL_FAST=<1|0>
+
+  If set to 1, --failfast is passed to the unittest framework to return upon
+  first failure.
+
+  Args:
+    argv: the argv to mutate in-place.
+  """
+
+  if argv is None:
+    return
+
+  if os.environ.get('TESTBRIDGE_TEST_RUNNER_FAIL_FAST') != '1':
+    return
+
+  argv[1:1] = ['--failfast']
+
+
 def _setup_sharding(custom_loader=None):
   # type: (Optional[unittest.TestLoader]) -> unittest.TestLoader
   """Implements the bazel sharding protocol.
@@ -2210,11 +2434,15 @@ def _setup_sharding(custom_loader=None):
 
   def getShardedTestCaseNames(testCaseClass):
     filtered_names = []
-    for testcase in sorted(delegate_get_names(testCaseClass)):
+    # We need to sort the list of tests in order to determine which tests this
+    # shard is responsible for; however, it's important to preserve the order
+    # returned by the base loader, e.g. in the case of randomized test ordering.
+    ordered_names = delegate_get_names(testCaseClass)
+    for testcase in sorted(ordered_names):
       bucket = next(bucket_iterator)
       if bucket == shard_index:
         filtered_names.append(testcase)
-    return filtered_names
+    return [x for x in ordered_names if x in filtered_names]
 
   base_loader.getTestCaseNames = getShardedTestCaseNames
   return base_loader
@@ -2228,6 +2456,8 @@ def _run_and_get_tests_result(argv, args, kwargs, xml_test_runner_class):
 
   # Set up test filtering if requested in environment.
   _setup_filtering(argv)
+  # Set up --failfast as requested in environment
+  _setup_test_runner_fail_fast(argv)
 
   # Shard the (default or custom) loader if sharding is turned on.
   kwargs['testLoader'] = _setup_sharding(kwargs.get('testLoader', None))
@@ -2272,6 +2502,14 @@ def _run_and_get_tests_result(argv, args, kwargs, xml_test_runner_class):
     # with partial information, in case the test process crashes).
     xml_buffer = six.StringIO()
     kwargs['testRunner'].set_default_xml_stream(xml_buffer)  # pytype: disable=attribute-error
+
+    # If we've used a seed to randomize test case ordering, we want to record it
+    # as a top-level attribute in the `testsuites` section of the XML output.
+    randomize_ordering_seed = getattr(
+        kwargs['testLoader'], '_randomize_ordering_seed', None)
+    setter = getattr(kwargs['testRunner'], 'set_testsuites_property', None)
+    if randomize_ordering_seed and setter:
+      setter('test_randomize_ordering_seed', randomize_ordering_seed)
   elif kwargs.get('testRunner') is None:
     kwargs['testRunner'] = _pretty_print_reporter.TextTestRunner
 
@@ -2288,9 +2526,9 @@ def _run_and_get_tests_result(argv, args, kwargs, xml_test_runner_class):
       runner.run_for_debugging = True
 
   # Make sure tmpdir exists.
-  if not os.path.isdir(FLAGS.test_tmpdir):
+  if not os.path.isdir(TEST_TMPDIR.value):
     try:
-      os.makedirs(FLAGS.test_tmpdir)
+      os.makedirs(TEST_TMPDIR.value)
     except OSError as e:
       # Concurrent test might have created the directory.
       if e.errno != errno.EEXIST:
@@ -2347,7 +2585,10 @@ def _get_qualname(cls):
 def _rmtree_ignore_errors(path):
   # type: (Text) -> None
   if os.path.isfile(path):
-    os.unlink(path)
+    try:
+      os.unlink(path)
+    except OSError:
+      pass
   else:
     shutil.rmtree(path, ignore_errors=True)
 
